/*
 * File System Interface
 *
 * myfs.c
 *
 * Implementation of the SNFS programming interface simulating the 
 * standard Unix I/O interface. This interface uses the SNFS API to
 * invoke the SNFS services in a remote server.
 * 
 */

#define _XOPEN_SOURCE 600

#include <myfs.h>
#include <snfs_api.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#ifndef SERVER_SOCK
#define SERVER_SOCK "/tmp/server.socket"
#endif


#define dprintf if (1) printf


// table number of file descriptors
#define NUM_FILE_DESC 10

typedef struct {
  int fileId;
  snfs_fhandle_t fhandle;
  unsigned fsize;
  unsigned offset;
} fd_t;


/*
 * Internal private global variables
 */

// file descriptors table
static fd_t fd_table[NUM_FILE_DESC];


/*
 * File descriptors management macros and functions
 */


static void fd_init()
{
  // set all file descriptors as unused
  for (int i = 0; i < NUM_FILE_DESC; i++) {
    fd_table[i].fileId = -1;
  }
}


static fd_t* fd_open()
{
  // find first unused file descriptor
  for (int i = 0; i < NUM_FILE_DESC; i++) {
    if (fd_table[i].fileId == -1) {
      fd_table[i].fileId = i;
      fd_table[i].offset = 0;
      return &fd_table[i];
    }
  }
  return NULL;
}


static fd_t * fd_lookup(int fileId)
{
  if (fileId < 0 || fileId >= NUM_FILE_DESC
      || fd_table[fileId].fileId != fileId) {
    return NULL;
  }
  return &fd_table[fileId];
}


static int fd_close(int fileId)
{
  if (fileId < 0 || fileId >= NUM_FILE_DESC
      || fd_table[fileId].fileId != fileId) {
    return -1;
  }
  // set file descriptor as unused
  fd_table[fileId].fileId = -1;
  return 0;
}


/*
 * Internal private auxiliary functions
 */


// file type
typedef enum {MY_DIR = 1, MY_FILE = 2} my_type_t;


static int my_get_dir(char* path, snfs_fhandle_t *dir, my_type_t type) 
{
  if (path[0] != '/' || path[1] == '/') {
    dprintf("[my_get_dir] invalid path.\n");
    return -1;
  }

  int i;
  for (i = 1; path[i] != '\0' && path[i] != '/'; i++);
  
  if (path[i] == '\0') { /* file */
    if (type == MY_DIR && i > 1) {
      unsigned fsize;
      if (snfs_lookup(*dir, path + 1, dir, &fsize) != STAT_OK) {
	dprintf("[my_get_dir] invalid directory: %s.\n", path);
	return -1;
      }
    }
    return 0;
  } else { /* dir */
    char dirname[MAX_FILE_NAME_SIZE];
    strncpy(dirname, path, i);
    dirname[i] = '\0';
    unsigned fsize;
    
    if (snfs_lookup(*dir, dirname + 1, dir, &fsize) != STAT_OK) {
      dprintf("[my_get_dir] invalid directory: %s.\n", dirname);
      return -1;
    }
    
    return my_get_dir(path + i, dir, type);
  }
}


/*
 * MYFS API implementation (see myfs.h)
 */


int my_init_lib()
{
  char client_sock[] = "/tmp/client.socket.XXXXXX";

  if(mkstemp(client_sock) == -1) {
    dprintf("[my_init_lib] unable to generate an unique socket name.\n");
    return -1;
  }

  if (snfs_init(client_sock, SERVER_SOCK) < 0) {
    dprintf("[my_init_lib] unable to initialize myfs API.\n");
    return -1;
  }
  
  // initialize file descriptors table
  fd_init();
  return 0;
}


int my_open(char* nome, int flags)
{
  if (nome == NULL) {
    dprintf("[my_open] malformed arguments.\n");
    return -1;
  }

  fd_t *fd = fd_open();
  if (fd == NULL) {
    dprintf("[my_open] maximum number of files opened.\n");
    return -1;
  }

  // get directory of the file
  snfs_fhandle_t dir = ROOT_FHANDLE;
  if (my_get_dir(nome, &dir, MY_FILE) != 0) {
    return -1;
  }

  // format request
  int i;
  for (i = strlen(nome) - 1; nome[i] != '/'; i--);

  if (snfs_lookup(dir, nome + i + 1, &fd->fhandle, &fd->fsize) != STAT_OK) {
    // checks if we want to create the file
    if (flags != O_CREATE) {
      dprintf("[my_open] cannot open file.\n");
      return -1;
    } else if (snfs_create(dir, nome + i + 1, &fd->fhandle) != STAT_OK) {
      dprintf("[my_open] cannot create file.\n");
      return -1;
    }
  }
  return fd->fileId;
}


int my_read(int fileId, char* buffer, unsigned numBytes)
{
  if (buffer == NULL) {
    dprintf("[my_read] malformed arguments.\n");
    return -1;
  }

  fd_t *fd = fd_lookup(fileId);
  if (fd == NULL) {
    dprintf("[my_read] invalid file id.\n");
    return -1;
  }

  // check if we want to read more than the size of the file left
  if (fd->offset + numBytes > fd->fsize) {
    numBytes = fd->fsize - fd->offset;
  }

  int numRead = 0;
  while(numBytes != 0) {
    // format request
    int nread = 0;
    unsigned nbytes = (numBytes < MAX_WRITE_DATA) ? numBytes : MAX_WRITE_DATA;
    numBytes -= nbytes;

    if (snfs_read(fd->fhandle, fd->offset, nbytes, buffer, &nread)
	!= STAT_OK) {
      dprintf("[my_read] data corrupted.\n");
      return -1;
    }

    // updates the file descriptor data and the number of bytes read
    fd->offset += nread;
    buffer += nread;
    numRead += nread;
  }
  return numRead;
}


int my_write(int fileId, char* buffer, unsigned numBytes)
{
  if (buffer == NULL) {
    dprintf("[my_write] malformed arguments.\n");
    return -1;
  }

  fd_t *fd = fd_lookup(fileId);
  if (fd == NULL) {
    dprintf("[my_write] invalid file id.\n");
    return -1;
  }

  int numWrite = 0;
  while(numBytes != 0) {
    // format request
    unsigned nbytes = (numBytes < MAX_READ_DATA) ? numBytes : MAX_READ_DATA;
    numBytes -= nbytes;

    if (snfs_write(fd->fhandle, fd->offset, nbytes, buffer, &fd->fsize)
	!= STAT_OK) {
      dprintf("[my_write] data corrupted.\n");
      return -1;
    }

    // updates the file descriptor data and the number of bytes written
    fd->offset += nbytes;
    buffer += nbytes;
    numWrite += nbytes;
  }
  return numWrite;
}


int my_close(int fileId)
{
  // update the file descriptor
  return fd_close(fileId);
}


int my_listdir(char* path, char **filenames, int* numFiles)
{
  if (path == NULL || filenames == NULL || numFiles == NULL) {
    dprintf("[my_listdir] malformed arguments.\n");
    return -1;
  }

  snfs_fhandle_t dir = ROOT_FHANDLE;  
  if (my_get_dir(path, &dir, MY_DIR) != 0) {
    dprintf("[my_listdir] invalid directory.\n"); 
    return -1;
  }

  // format request
  unsigned cmax = 64, count, len = 0;
  snfs_dir_entry_t list[MAX_READDIR_ENTRIES];

  if (snfs_readdir(dir, cmax, list, &count) != STAT_OK) {
    dprintf("[my_listdir] error retrieving directory list.\n");
    return -1;
  }

  // calculate the space of filenames (not including '\0')
  for (int i = 0; i < count; i++) {
    len += list[i].len;
  }

  // allocate space to write the filenames
  *filenames = (char*) malloc(len + count);
  for (int i = 0, len = 0; i < count; i++) {
    strncpy(&(*filenames)[len], list[i].name, list[i].len + 1);
    len += list[i].len + 1;
  }

  *numFiles = count;
  return 0;
}


int my_mkdir(char *dirname)
{
  if (dirname == NULL) {
    dprintf("[my_mkdir] malformed arguments.\n");
    return -1;
  }

  snfs_fhandle_t dir = ROOT_FHANDLE;  
  if (my_get_dir(dirname, &dir, MY_FILE) != 0) {
    dprintf("[my_mkdir] invalid directory.\n"); 
    return -1;
  }

  // format request
  int i;
  for (i = strlen(dirname) - 1; dirname[i] != '/'; i--);

  snfs_fhandle_t newdir;
  if (snfs_mkdir(dir, dirname + i + 1, &newdir) != STAT_OK) {
    dprintf("[my_mkdir] error creating directory.\n");
    return -1;
  }

  return 0;
}
